use serde::Serialize;
use rocket::http::Status;
use rocket::request::{Request, Outcome};
use rocket::response::{self, Response, Responder};
use rocket_contrib::json::Json;
use serde_json::Value;
use djangohashers::{HasherError};
use diesel::result::Error as DieselError;
use crate::dns::dns_api::DnsApiError;
use crate::models;

#[derive(Debug)]
pub enum UserError {
    ZoneNotFound,
    NotFound,
    UserConflict,
    BadCreds,
    BadToken,
    ExpiredToken,
    MalformedHeader,
    PermissionDenied,
    DbError(DieselError),
    PasswordError(HasherError),
}

impl From<HasherError> for UserError {
    fn from(e: HasherError) -> Self {
        UserError::PasswordError(e)
    }
}

impl From<DieselError> for UserError {
    fn from(e: DieselError) -> Self {
        UserError::DbError(e)
    }
}


#[derive(Serialize, Debug)]
pub struct ErrorResponse {
    #[serde(with = "StatusDef")]
    #[serde(flatten)]
    pub status: Status,
    pub message: String,
    #[serde(skip_serializing_if = "Option::is_none")]
    pub details: Option<Value>
}

#[derive(Serialize)]
#[serde(remote = "Status")]
struct StatusDef {
    code: u16,
    #[serde(rename = "status")]
    reason: &'static str,
}

impl ErrorResponse {
    pub fn new(status: Status, message: String) -> ErrorResponse {
        ErrorResponse {
            status,
            message,
            details: None,
        }
    }

    pub fn with_details<T: Serialize> (self, details: T) -> ErrorResponse {
        ErrorResponse {
            details: serde_json::to_value(details).ok(),
            ..self
        }
    }

    pub fn err<R>(self) -> Result<R, ErrorResponse> {
        Err(self)
    }
}

impl<'r> Responder<'r, 'static> for ErrorResponse {
    fn respond_to(self, req: &'r Request<'_>) -> response::Result<'static> {
        let status = self.status;
        Response::build_from(Json(self).respond_to(req)?).status(status).ok()
    }
}

impl From<UserError> for ErrorResponse {
    fn from(e: UserError) -> Self {
        match e {
            UserError::BadCreds => ErrorResponse::new(Status::Unauthorized, "Provided credentials or token do not match any existing user".into()),
            UserError::UserConflict => ErrorResponse::new(Status::Conflict, "This user already exists".into()),
            UserError::NotFound => ErrorResponse::new(Status::NotFound, "User does not exist".into()),
            UserError::BadToken => ErrorResponse::new(Status::BadRequest, "Malformed token".into()),
            UserError::ExpiredToken => ErrorResponse::new(Status::Unauthorized, "The provided token has expired".into()),
            UserError::MalformedHeader => ErrorResponse::new(Status::BadRequest, "Malformed authorization header".into()),
            UserError::PermissionDenied => ErrorResponse::new(Status::Forbidden, "Bearer is not authorized to access the resource".into()),
            UserError::ZoneNotFound => ErrorResponse::new(Status::NotFound, "DNS zone does not exist".into()),
            UserError::DbError(e) => make_500(e),
            UserError::PasswordError(e) => make_500(e)
        }
    }
}

impl From<DnsApiError> for ErrorResponse {
    fn from(e: DnsApiError) -> Self {
        match e {
            DnsApiError::ResponceNotOk { code, zone } => {
                println!("Query for zone {} failed with code {}", zone, code);

                ErrorResponse::new(
                    Status::NotFound,
                    "Zone could not be found".into()
                ).with_details(json!({
                    "zone_name": zone.to_utf8()
                }))
            },
            DnsApiError::ClientError(e) => make_500(e)
        }
    }
}

impl From<models::RecordListParseError> for ErrorResponse {
    fn from(e: models::RecordListParseError) -> Self {
        match e {
            models::RecordListParseError::RecordNotInZone { zone, class, mismatched_class, mismatched_zone} => {
                ErrorResponse::new(
                    Status::BadRequest,
                    "Record list contains records that do not belong to the zone".into()
                ).with_details(
                    json!({
                        "zone_name": zone.to_utf8(),
                        "class": models::DNSClass::from(class),
                        "mismatched_class": mismatched_class,
                        "mismatched_zone": mismatched_zone,
                    })
                )
            },
            models::RecordListParseError::ParseError { zone, bad_records } => {
                ErrorResponse::new(
                    Status::BadRequest,
                    "Record list contains records that could not be parsed into DNS records".into()
                ).with_details(
                    json!({
                        "zone_name": zone.to_utf8(),
                        "records": bad_records
                    })
                )
            }
        }
    }
}


impl<S> From<ErrorResponse> for Outcome<S, ErrorResponse> {
    fn from(e: ErrorResponse) -> Self {
        Outcome::Failure(e.into())
    }
}


impl From<ErrorResponse> for (Status, ErrorResponse) {
    fn from(e: ErrorResponse) -> Self {
        (e.status, e)
    }
}

pub fn make_500<E: std::fmt::Debug>(e: E) -> ErrorResponse {
    println!("Making 500 for Error: {:?}", e);
    ErrorResponse::new(Status::InternalServerError, "An unexpected error occured.".into())
}
